import os
import io
from typing import Union, List, Tuple

import requests
import json
import base64
import urllib3
from urllib.parse import quote
from pathlib import Path

from accli.common import todict

ACCLI_DEBUG = os.environ.get('ACCLI_DEBUG', False)


class AccAPIError(Exception):
    pass

    # def __init__(self, *args, **kwargs):
    #     self.status_code = kwargs.pop('status_code')
    #     self.response_data = kwargs.pop('response_data')
    #     super().__init__(*args, **kwargs)


retries = urllib3.util.Retry(total=10, backoff_factor=1)

http_client = urllib3.poolmanager.PoolManager(
    num_pools=20, retries=retries, maxsize=2000, block=True
)

http_client_wo_cert_verification = urllib3.poolmanager.PoolManager(
    cert_reqs="CERT_NONE", num_pools=20, retries=retries, maxsize=2000, block=True
)


class AcceleratorJobProjectService:
    def __init__(
            self,
            user_token,
            server_url='http://accelerator.iiasa.ac.at',
            verify_cert=(not ACCLI_DEBUG)
    ):

        self.user_token = user_token

        if verify_cert:
            self.http_client = http_client
        else:
            self.http_client = http_client_wo_cert_verification

        self.cli_base_url = f"{server_url}/api/v1/ajob-cli"
        self.common_request_headers = {
            'x-authorization': user_token
        }

    def http_client_request(self, *args, **kwargs):

        if urllib3.__version__.startswith('1.'):
            if 'json' in kwargs:
                json_dict = kwargs.pop('json')

                assert isinstance(json_dict, dict), (
                    f"urllib request: json keyword argument should be dict."
                )

                encoded_data = json.dumps(json_dict).encode('utf-8')

                kwargs['body'] = encoded_data

        res = self.http_client.request(*args, **kwargs)

        if str(res.status)[0] in ['4', '5']:
            raise AccAPIError(
                f"Accelerator api error:: status_code={res.status} :: response_data={res.data}",
            )
        return res

    def get_file_stat(self, bucket_object_id):
        res = self.http_client_request(
            "GET",
            f"{self.cli_base_url}/file-stat/{bucket_object_id}/",
            headers=self.common_request_headers
        )
        return todict(res.data)

    def enumerate_files_by_prefix(self, prefix):
        raise NotImplementedError("filename parameter is not implemented yet")

        project_slug = filename.split('/')[0]

        b64_encoded_prefix = base64.b64encode(prefix.encode()).decode()

        res = self.http_client_request(
            "GET",
            f"{self.cli_base_url}/{project_slug}/enumerate-all-files/{b64_encoded_prefix}/",
            headers=self.common_request_headers
        )
        if res.data:
            return todict(res.data)

    def get_file_url_from_repo(self, filename):
        project_slug = filename.split('/')[0]
        res = self.http_client_request(
            "GET",
            f"{self.cli_base_url}/{project_slug}/get-file-download-url/?filename={filename}",
            headers=self.common_request_headers
        )
        if res.data:
            return todict(res.data)
        return None

    def get_dataset_type(self, *args, **kwargs):
        return self.get_bucket_object_validation_type(*args, **kwargs)

    def get_bucket_object_validation_type(self, bucket_object_id):
        res = self.http_client_request(
            "GET",
            f"{self.cli_base_url}/dataset-type/{bucket_object_id}/",
            headers=self.common_request_headers
        )
        if res.data:
            return todict(res.data)
        return None

    def get_filename_dataset_type(self, filename):
        res = self.http_client_request(
            "GET",
            f"{self.cli_base_url}/filename-dataset-type/?filename={filename}",
            headers=self.common_request_headers
        )
        if res.data:
            return todict(res.data)
        return None

    def get_filename_validation_details(self, filename):
        res = self.http_client_request(
            "GET",
            f"{self.cli_base_url}/filename-validation-detail/?filename={filename}",
            headers=self.common_request_headers
        )
        if res.data:
            return todict(res.data)
        return None

    def get_bucket_object_validation_details(self, bucket_object_id):
        res = self.http_client_request(
            "GET",
            f"{self.cli_base_url}/validation-detail/{bucket_object_id}/",
            headers=self.common_request_headers
        )
        if res.data:
            return todict(res.data)
        return None

    def get_file_url(self, bucket_object_id):
        res = self.http_client_request(
            "GET",
            f"{self.cli_base_url}/file-url/{bucket_object_id}/",
            headers=self.common_request_headers
        )

        if res.data:
            return todict(res.data)
        return None

    def get_file_stream(self, bucket_object_id):
        url = self.get_file_url(bucket_object_id)
        if url:
            resp = self.http_client_request("GET", url, preload_content=False)
            return resp
        return None

    def check_job_health(self):

        res = self.http_client_request(
            "GET",
            f"{self.cli_base_url}/is-healthy/",
            headers=self.common_request_headers
        )

        if res.data:
            res = todict(res.data)

        is_healthy = res['is_healthy']
        return is_healthy

    def add_log_file(self, data: bytes, filename):

        res = self.http_client_request(
            "GET",
            f"{self.cli_base_url}/presigned-log-upload-url/?filename={filename}",
            headers=self.common_request_headers
        )

        if res.data:
            res = todict(res.data)
        else:
            raise ValueError(f"Unable to get presign url in response: {res.data}")

        upload_url = res['upload_url']
        app_bucket_id = res['app_bucket_id']
        res_filename = res['filename']
        is_healthy = res['is_healthy']

        requests.put(
            upload_url,
            data=data,
            # headers=headers,
            verify=False,
        )

        self.http_client_request(
            "POST",
            f"{self.cli_base_url}/register-log-file/",
            json=dict(
                filename=res_filename,
                app_bucket_id=app_bucket_id
            ),
            headers=self.common_request_headers
        )

        return is_healthy

    def get_multipart_put_create_signed_url(
            self,
            app_bucket_id,
            object_name,
            upload_id,
            part_number
    ):
        res = self.http_client_request(
            "GET",
            f"{self.cli_base_url}/put-create-signed-url/",
            fields=dict(
                app_bucket_id=app_bucket_id,
                object_name=object_name,
                upload_id=upload_id,
                part_number=part_number
            ),
            headers=self.common_request_headers
        )

        return todict(res.data)

    def get_multipart_put_update_signed_url(
            self,
            filename,
            upload_id,
            part_number,
    ):
        res = self.http_client_request(
            "GET",
            f"{self.cli_base_url}/put-update-signed-url/",
            fields=dict(
                filename=filename,
                upload_id=upload_id,
                part_number=part_number
            ),
            headers=self.common_request_headers
        )

        return todict(res.data)

    def get_put_create_multipart_upload_id(self, filename):
        encoded_filename = quote(filename)
        res = self.http_client_request(
            "GET",
            f"{self.cli_base_url}/multipart-upload-id/?filename={encoded_filename}",
            headers=self.common_request_headers
        )

        data = todict(res.data)

        return data['upload_id'], data['app_bucket_id'], data['uniqified_filename']

    def get_validator_create_multipart_upload_id(self, filename):
        encoded_filename = quote(filename)
        res = self.http_client_request(
            "GET",
            f"{self.cli_base_url}/validator-multipart-upload-id/?filename={encoded_filename}",
            headers=self.common_request_headers
        )

        data = todict(res.data)

        return data['upload_id'], data['app_bucket_id'], data['uniqified_filename']

    def get_put_update_multipart_upload_id(self, filename):
        encoded_filename = quote(filename)
        res = self.http_client_request(
            "GET",
            f"{self.cli_base_url}/update-multipart-upload-id/?filename={encoded_filename}",
            headers=self.common_request_headers
        )

        return todict(res.data)

    def complete_job_multipart_upload(
            self,
            app_bucket_id,
            filename,
            upload_id,
            parts: List[Tuple[str, str]],
            is_log_file=False
    ):
        headers = {"Content-Type": "application/json"}

        headers.update(self.common_request_headers)

        res = self.http_client_request(
            "PUT",
            f"{self.cli_base_url}/complete-create-multipart-upload/",
            json=dict(
                app_bucket_id=app_bucket_id,
                filename=filename,
                upload_id=upload_id,
                parts=base64.b64encode(json.dumps(parts).encode()).decode(),
                is_log_file=is_log_file
            ),
            headers=headers
        )

        return todict(res.data)

    def complete_validator_multipart_upload(
            self,
            app_bucket_id,
            filename,
            upload_id,
            parts: List[Tuple[str, str]]
    ):
        headers = {"Content-Type": "application/json"}

        headers.update(self.common_request_headers)

        res = self.http_client_request(
            "PUT",
            f"{self.cli_base_url}/complete-validator-multipart-upload/",
            json=dict(
                app_bucket_id=app_bucket_id,
                filename=filename,
                upload_id=upload_id,
                parts=base64.b64encode(json.dumps(parts).encode()).decode()
            ),
            headers=headers
        )

        return todict(res.data)

    def complete_update_multipart_upload(
            self, filename, upload_id, parts: List[Tuple[str, str]]
    ):

        headers = {"Content-Type": "application/json"}

        headers.update(self.common_request_headers)

        res = self.http_client_request(
            "PUT",
            f"{self.cli_base_url}/complete-update-multipart-upload/",
            json=dict(
                filename=filename,
                upload_id=upload_id,
                parts=base64.b64encode(json.dumps(parts).encode()).decode()
            ),
            headers=headers
        )

        return todict(res.data)

    def abort_create_multipart_upload(self, app_bucket_id, filename, upload_id):

        headers = {"Content-Type": "application/json"}

        headers.update(self.common_request_headers)

        res = self.http_client_request(
            "PUT",
            f"{self.cli_base_url}/abort-create-multipart-upload/",
            json=dict(
                app_bucket_id=app_bucket_id,
                filename=filename,
                upload_id=upload_id
            ),
            headers=headers
        )

    def abort_update_multipart_upload(self, filename, upload_id):

        headers = {
            # "Content-Type": "application/json"
        }

        headers.update(self.common_request_headers)

        res = self.http_client_request(
            "PUT",
            f"{self.cli_base_url}/abort-update-multipart-upload/",
            json=dict(
                filename=filename,
                upload_id=upload_id
            ),
            headers=headers
        )

    # # TODO @wrufesh ensure it requires special token
    # def register_iamc_validation(
    #     self, validated_bucket_object_id, indexdb_bucket_object_id
    # ):

    #     headers = {"Content-Type": "application/json"}

    #     headers.update(self.common_request_headers)

    #     res = self.http_client_request(
    #         "PUT", 
    #         f"{self.cli_base_url}/register-iamc-validation/",
    #         json=dict(
    #             validated_bucket_object_id=validated_bucket_object_id,
    #             indexdb_bucket_object_id=indexdb_bucket_object_id
    #         ),
    #         headers=headers
    #     )

    def register_validation(
            self,
            validated_bucket_object_id: int,
            dataset_template_id: int,
            validated_metadata: dict,
            validation_supporting_bucket_object_ids: List[int]
    ):
        headers = {"Content-Type": "application/json"}

        headers.update(self.common_request_headers)

        res = self.http_client_request(
            "PUT",
            f"{self.cli_base_url}/register-validation/",
            json=dict(
                validated_bucket_object_id=validated_bucket_object_id,
                dataset_template_id=dataset_template_id,
                validated_metadata=validated_metadata,
                validation_supporting_bucket_object_ids=validation_supporting_bucket_object_ids
            ),
            headers=headers
        )

    def get_dataset_template_details(
            self,
            dataset_template_id
    ):
        headers = {"Content-Type": "application/json"}

        headers.update(self.common_request_headers)

        res = self.http_client_request(
            "GET",
            f"{self.cli_base_url}/dataset-template-detail/{dataset_template_id}/",
            headers=headers
        )

        return todict(res.data)

    def read_part_data(self, stream, size, part_data=b"", progress=None):
        """Read part data of the given size from the stream."""
        size -= len(part_data)
        while size:
            data = stream.read(size)
            if not data:
                break  # EOF reached
            if not isinstance(data, bytes):
                raise ValueError("read() must return 'bytes' object")
            part_data += data
            size -= len(data)
            if progress:
                progress.update(len(data))
        return part_data

    def add_filestream_as_job_output(self, filename, file_stream, is_log_file=False):
        headers = dict()
        headers["Content-Type"] = "application/octet-stream"

        part_size, part_count = 50 * 1024 ** 2, -1

        upload_id = None
        app_bucket_id = None
        uniqified_filename = None

        one_byte = b""
        stop = False
        part_number = 0
        parts = []
        uploaded_size = 0
        put_presigned_url = None

        try:
            while not stop:
                part_number += 1
                part_data = self.read_part_data(
                    file_stream,
                    part_size + 1,
                    one_byte,
                    progress=None,
                )

                # If part_data_size is less or equal to part_size,
                # then we have reached the last part.
                if len(part_data) <= part_size:
                    part_count = part_number
                    stop = True
                else:
                    one_byte = part_data[-1:]
                    part_data = part_data[:-1]

                uploaded_size += len(part_data)

                if not upload_id:
                    (
                        upload_id,
                        app_bucket_id,
                        uniqified_filename,
                    ) = self.get_put_create_multipart_upload_id(
                        filename,
                        # headers=headers
                    )

                put_presigned_url = self.get_multipart_put_create_signed_url(
                    app_bucket_id, uniqified_filename, upload_id, part_number
                )

                part_upload_response = requests.put(
                    put_presigned_url,
                    data=part_data,
                    # headers=headers,
                    verify=False,
                )

                etag = part_upload_response.headers.get("etag").replace('"', "")
                parts.append((part_number, etag))

            created_bucket_object_id = self.complete_job_multipart_upload(
                app_bucket_id, uniqified_filename, upload_id, parts, is_log_file=is_log_file
            )
            return created_bucket_object_id

        except Exception as err:
            # Cancel if any error
            if upload_id:
                self.abort_create_multipart_upload(
                    app_bucket_id,
                    uniqified_filename,
                    upload_id,
                )
            raise err

    def replace_bucket_object_id_content(self, filename, file_stream):
        headers = dict()
        headers["Content-Type"] = "application/octet-stream"

        part_size, part_count = 50 * 1024 ** 2, -1

        upload_id = None
        app_bucket_id = None
        uniqified_filename = None

        one_byte = b""
        stop = False
        part_number = 0
        parts = []
        uploaded_size = 0
        put_presigned_url = None

        try:
            while not stop:
                part_number += 1
                part_data = self.read_part_data(
                    file_stream,
                    part_size + 1,
                    one_byte,
                    progress=None,
                )

                # If part_data_size is less or equal to part_size,
                # then we have reached the last part.
                if len(part_data) <= part_size:
                    part_count = part_number
                    stop = True
                else:
                    one_byte = part_data[-1:]
                    part_data = part_data[:-1]

                uploaded_size += len(part_data)

                if not upload_id:
                    upload_id = self.get_put_update_multipart_upload_id(
                        filename
                    )

                put_presigned_url = self.get_multipart_put_update_signed_url(
                    filename, upload_id, part_number
                )

                part_upload_response = requests.put(
                    put_presigned_url,
                    data=part_data,
                    # headers=headers,
                    verify=False,
                )

                etag = part_upload_response.headers.get("etag").replace('"', "")
                parts.append((part_number, etag))

            created_bucket_object_id = self.complete_update_multipart_upload(
                filename,
                upload_id,
                parts,
            )

            return created_bucket_object_id
        except Exception as err:
            # Cancel if any error
            if upload_id:
                self.abort_update_multipart_upload(
                    filename,
                    upload_id,
                )

            raise err

    def add_filestream_as_validation_supporter(self, filename, file_stream, is_log_file=False):
        headers = dict()
        headers["Content-Type"] = "application/octet-stream"

        part_size, part_count = 50 * 1024 ** 2, -1

        upload_id = None
        app_bucket_id = None
        uniqified_filename = None

        one_byte = b""
        stop = False
        part_number = 0
        parts = []
        uploaded_size = 0
        put_presigned_url = None

        try:
            while not stop:
                part_number += 1
                part_data = self.read_part_data(
                    file_stream,
                    part_size + 1,
                    one_byte,
                    progress=None,
                )

                # If part_data_size is less or equal to part_size,
                # then we have reached the last part.
                if len(part_data) <= part_size:
                    part_count = part_number
                    stop = True
                else:
                    one_byte = part_data[-1:]
                    part_data = part_data[:-1]

                uploaded_size += len(part_data)

                if not upload_id:
                    (
                        upload_id,
                        app_bucket_id,
                        uniqified_filename,
                    ) = self.get_validator_create_multipart_upload_id(
                        filename,
                        # headers=headers
                    )

                put_presigned_url = self.get_multipart_put_create_signed_url(
                    app_bucket_id, uniqified_filename, upload_id, part_number
                )

                part_upload_response = requests.put(
                    put_presigned_url,
                    data=part_data,
                    # headers=headers,
                    verify=False,
                )

                etag = part_upload_response.headers.get("etag").replace('"', "")
                parts.append((part_number, etag))

            created_bucket_object_id = self.complete_validator_multipart_upload(
                app_bucket_id, uniqified_filename, upload_id, parts
            )
            return created_bucket_object_id

        except Exception as err:
            # Cancel if any error
            if upload_id:
                self.abort_create_multipart_upload(
                    app_bucket_id,
                    uniqified_filename,
                    upload_id,
                )
            raise err

    def update_job_status(self, status):
        headers = {
            # "Content-Type": "application/json"
            # 'Content-Type': 'application/x-www-form-urlencoded'
        }

        headers.update(self.common_request_headers)

        res = self.http_client_request(
            "POST",
            f"{self.cli_base_url}/webhook-event/",
            json=dict(
                type='STATUS_UPDATE',
                data=dict(
                    new_status=status
                )
            ),
            headers=headers
        )


class Fs:
    @staticmethod
    def write_stream_remote(filestream, dest_filepath, user_token, server_url):

        accelerator_job_service = AcceleratorJobProjectService(
            user_token,
            server_url=server_url,
            verify_cert=(not ACCLI_DEBUG)
        )

        accelerator_job_service.add_filestream_as_job_output(
            dest_filepath,
            filestream
        )

    @staticmethod
    def write_stream_local(filestream, dest_filepath):
        with open(dest_filepath, 'wb') as file:
            file.write(filestream.getvalue())

    @staticmethod
    def get_file_url(remote_filepath):
        user_token = os.environ.get("ACC_JOB_TOKEN", None)
        server_url = os.environ.get("ACC_JOB_GATEWAY_SERVER", None)

        if not (user_token and server_url):
            raise ValueError("Remote data repository credentials not found.")

        accelerator_job_service = AcceleratorJobProjectService(
            user_token,
            server_url=server_url,
            verify_cert=(not ACCLI_DEBUG)
        )

        return accelerator_job_service.get_file_url_from_repo(
            remote_filepath
        )

    @staticmethod
    def write_file(source: Union[str, io.BytesIO], dest_filepath):
        user_token = os.environ.get("ACC_JOB_TOKEN", None)
        server_url = os.environ.get("ACC_JOB_GATEWAY_SERVER", None)

        if isinstance(source, io.BytesIO):
            if user_token and server_url:
                Fs.write_stream_remote(source, dest_filepath, user_token, server_url)
            else:
                Fs.write_stream_local(source, dest_filepath)
        else:
            if type(source) != str:
                raise ValueError("Unknown source argument")
            source_path = Path(source)
            if not source_path.exists():
                raise ValueError("Source path does not exist")

            with open(source, 'rb') as fstream:
                if user_token and server_url:
                    Fs.write_stream_remote(fstream, dest_filepath, user_token, server_url)
                else:
                    Fs.write_stream_local(fstream, dest_filepath)

    @staticmethod
    def enumerate_files_by_prefix(prefix):
        user_token = os.environ.get("ACC_JOB_TOKEN", None)
        server_url = os.environ.get("ACC_JOB_GATEWAY_SERVER", None)

        if not (user_token and server_url):
            raise ValueError("Remote data repository credentials not found.")

        accelerator_job_service = AcceleratorJobProjectService(
            user_token,
            server_url=server_url,
            verify_cert=(not ACCLI_DEBUG)
        )

        return accelerator_job_service.enumerate_files_by_prefix(
            prefix
        )



import glob
import importlib.util
import os
import re
import warnings
from contextlib import contextmanager
from pathlib import Path

import requests
import typer
from typing import List

from accli.AcceleratorTerminalCliProjectService import AcceleratorTerminalCliProjectService
from accli.CsvRegionalTimeseriesValidator import CsvRegionalTimeseriesValidator
from accli.token import save_token_details, get_token, get_server_url, set_project_slug
from rich import print
from rich.progress import Progress, TextColumn, BarColumn, TaskProgressColumn, TimeElapsedColumn
from typing_extensions import Annotated

from ._version import VERSION

warnings.filterwarnings('ignore')

ACCLI_DEBUG = os.environ.get('ACCLI_DEBUG', False)

app = typer.Typer(
    add_completion=False,
    pretty_exceptions_show_locals=False,
    no_args_is_help=True
)


def get_size(path):
    size = 0

    for file in glob.iglob(f"{path}/**/*.*", recursive=True):
        size += os.path.getsize(file)

    return size


@contextmanager
def pushd(new_dir):
    """Temporarily change the working directory."""
    prev_dir = os.getcwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(prev_dir)


@app.command()
def about():
    print(
        "[bold cyan]This is a terminal client for Accelerator hosted on https://accelerator.iiasa.ac.at . [/bold cyan]\n")
    print(
        "[bold cyan]Please file feature requests and suggestions at https://github.com/iiasa/accli/issues .[/bold cyan]\n")
    print("[bold cyan]License: The MIT License (MIT)[/bold cyan]\n")
    print(f"[bold cyan]Version: {VERSION}[/bold cyan]\n")


@app.command()
def login(
        server: Annotated[
            str, typer.Option(..., '-s', help="Accelerator server url.")] = "https://accelerator.iiasa.ac.at",
        webcli: Annotated[str, typer.Option(..., '-c',
                                            help="Accelerator web client for authorization.")] = "https://accelerator.iiasa.ac.at"
):
    print(
        f"[bold cyan]Welcome to Accelerator Terminal Client.[/bold cyan]\n"
        f"[bold cyan]Powered by IIASA[/bold cyan]\n"
    )

    print(f"[italic]Get authorization code on following web url: [cyan]{webcli}/acli-auth-code[cyan][/italic] \n")

    device_authorization_code = typer.prompt("Enter the authorization code?")

    token_response = requests.post(
        f"{server}/api/v1/oauth/device/token/",
        json={"device_authorization_code": device_authorization_code},
        verify=(not ACCLI_DEBUG)
    )

    print("")

    if token_response.status_code == 400:
        print(f"[bold red]ERROR: {token_response.json().get('detail')}[/red]")
        raise typer.Exit(1)

    save_token_details(token_response.json(), server, webcli)

    print("[bold green]Successfully logged in.[/bold green]:rocket: :rocket:")


def upload_file(project_slug, accelerator_filename, local_filepath, progress, task, folder_name,
                max_workers=os.cpu_count()):
    access_token = get_token()

    server_url = get_server_url()

    term_cli_project_service = AcceleratorTerminalCliProjectService(
        user_token=access_token,
        server_url=server_url,
        verify_cert=(not ACCLI_DEBUG)
    )

    stat = term_cli_project_service.get_file_stat(project_slug, f"{folder_name}/{accelerator_filename}")

    if stat:
        progress.update(task, advance=stat.get('size'))
    else:

        with open(local_filepath, 'rb') as file_stream:
            term_cli_project_service.upload_filestream_to_accelerator(project_slug,
                                                                      f"{folder_name}/{accelerator_filename}",
                                                                      file_stream, progress, task,
                                                                      max_workers=max_workers)


@app.command()
def upload(
        project_slug: Annotated[str, typer.Argument(help="Unique Accelerator project slug.")],
        path: Annotated[str, typer.Argument(help="Folder path to upload to Accelerator project space.")],
        folder_name: Annotated[str, typer.Argument(help="Name of the folder to be made in Accelerator project space.")],
        max_workers: Annotated[
            int, typer.Option(..., '-w', help="Maximum worker pool for multipart upload.")] = os.cpu_count()
):
    # TODO make user free to put anywere except for reserved folder

    if not re.fullmatch(r'[a-zA-Z0-9\-_]+', folder_name):
        raise ValueError("Folder name is invalid.")

    with Progress(
            TextColumn("[progress.description]"),
            BarColumn(),
            TaskProgressColumn(),
            TimeElapsedColumn(),
            TextColumn("{task.description}"),
            transient=True
    ) as progress:

        if os.path.isdir(path):

            if not path.endswith("/"):
                path = path + "/"

            folder_size = get_size(path)
            print('Folder size', folder_size)
            upload_task = progress.add_task("[cyan]Uploading.", total=folder_size)

            for local_file_path in glob.iglob(f"{path}/**/*.*", recursive=True):
                accelerator_filename = os.path.relpath(local_file_path, path)

                if os.name == 'nt':
                    accelerator_filename = accelerator_filename.replace('\\', '/')

                progress.update(
                    upload_task,
                    description=f"[cyan]Uploading {local_file_path} \t"
                )

                if not os.path.isfile(local_file_path):
                    continue
                upload_file(project_slug, accelerator_filename, local_file_path, progress, upload_task, folder_name,
                            max_workers=max_workers)
        elif os.path.isfile(path):
            raise NotImplementedError('Only folder can be uploaded. File upload is not implemented.')
        else:
            print("ERROR: No such file or directory.")
            typer.Exit(1)


@app.command()
def validate(
        project_slug: Annotated[str, typer.Argument(help="Unique Accelerator project slug.")],
        template_slug: Annotated[str, typer.Argument(help="Unique project template slug")],
        filepath: Annotated[str, typer.Argument(help="Path of the file to validate")],
        server: Annotated[
            str, typer.Option(..., '-s', help="Accelerator server url.")] = "https://accelerator.iiasa.ac.at",
):
    term_cli_project_service = AcceleratorTerminalCliProjectService(
        user_token="",
        server_url=server,
        verify_cert=(not ACCLI_DEBUG)
    )

    validate = CsvRegionalTimeseriesValidator(
        project_slug=project_slug,
        dataset_template_slug=template_slug,
        input_filepath=filepath,
        project_service=term_cli_project_service,
    )

    validate()


@app.command()
def dispatch(
        project_slug: Annotated[str, typer.Argument(help="Unique Accelerator project slug.")],
        root_task_variable: Annotated[str, typer.Argument(help="Root task variable in workflow_file.")],
        workflow_filename: Annotated[str, typer.Option(..., '-f', help="Python workflow filepath.")] = "wkube.py"
):
    set_project_slug(project_slug)
    access_token = get_token()
    server_url = get_server_url()

    term_cli_project_service = AcceleratorTerminalCliProjectService(
        user_token=access_token,
        server_url=server_url,
        verify_cert=(not ACCLI_DEBUG)
    )

    # ✅ Resolve the workflow file path properly
    if os.path.isabs(workflow_filename):
        workflow_filepath = workflow_filename
    else:
        workflow_filepath = os.path.abspath(os.path.join(os.getcwd(), workflow_filename))

    if not os.path.isfile(workflow_filepath):
        raise FileNotFoundError(f"Workflow file not found: {workflow_filepath}")

    workflow_dir = os.path.dirname(workflow_filepath)

    # ✅ Temporarily switch to the workflow file's directory
    with pushd(workflow_dir):
        # ✅ Import the module dynamically
        spec = importlib.util.spec_from_file_location("workflow", workflow_filepath)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        job_to_dispatch = getattr(module, root_task_variable, None)

        if not job_to_dispatch:
            raise ValueError(f"No root task variable found with name '{root_task_variable}' in {workflow_filepath}")

        print(job_to_dispatch.description)

        root_job_id = term_cli_project_service.dispatch(
            project_slug,
            job_to_dispatch.description
        )

    print(f"Dispatched root job #ID: {root_job_id}")


@app.command()
def copy(
        acc_src: Annotated[str, typer.Argument(help="Source path in Accelerator project space")],
        destination: Annotated[str, typer.Option(..., "-d", help="Destination directory")] = "./",
        token_pass: Annotated[str, typer.Option(..., "-t", help="Destination directory")] = "",
):
    access_token = get_token()
    server_url = get_server_url()

    term_cli_project_service = AcceleratorTerminalCliProjectService(
        user_token=access_token,
        server_url=server_url,
        verify_cert=(not ACCLI_DEBUG),
    )

    dest_path = Path(destination).expanduser().resolve()
    dest_path.mkdir(parents=True, exist_ok=True)

    filenames: List[str] = term_cli_project_service.enumerate_files_by_prefix(acc_src, token_pass=token_pass)

    with Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            TimeElapsedColumn(),
            transient=True,
    ) as progress:

        for filename in filenames:
            local_file = dest_path / filename
            local_file.parent.mkdir(parents=True, exist_ok=True)

            final_file = local_file
            partial_file = local_file.with_suffix(local_file.suffix + ".part")

            if final_file.exists():
                typer.echo(f"Skipping {final_file} (already exists)")
                continue

            typer.echo(f"Downloading {filename} -> {final_file}")
            try:
                file_url = term_cli_project_service.get_file_url_from_repo(filename, token_pass=token_pass)
                with requests.get(file_url, stream=True, verify=(not ACCLI_DEBUG)) as r:
                    r.raise_for_status()
                    total = int(r.headers.get("Content-Length", 0))

                    task = progress.add_task(
                        f"[cyan]Downloading {filename}", total=total or None
                    )

                    with open(partial_file, "wb") as f:
                        for chunk in r.iter_content(chunk_size=8192):
                            if chunk:
                                f.write(chunk)
                                progress.update(task, advance=len(chunk))

                partial_file.rename(final_file)
                typer.echo(f"✔ Downloaded {final_file}")

            except Exception as e:
                if partial_file.exists():
                    partial_file.unlink()
                typer.echo(f"✖ Failed to download {filename}: {e}", err=True)



import csv
import typing
from typing import Optional, List

import typer
from jsonschema import validate as jsonschema_validate
from jsonschema.exceptions import ValidationError, SchemaError
from rich import print
from rich.progress import Progress, SpinnerColumn, TextColumn

if typing.TYPE_CHECKING:
    from accli.AcceleratorTerminalCliProjectService import AcceleratorTerminalCliProjectService


def lower_rows(iterator):
    # return itertools.chain([next(iterator).lower()], iterator)
    for item in iterator:
        yield item.lower()


class CsvRegionalTimeseriesValidator:
    def __init__(
            self,
            *,
            project_slug,
            dataset_template_slug,
            input_filepath,
            project_service: 'AcceleratorTerminalCliProjectService',
            csv_fieldnames: Optional[List[str]] = None,

    ):
        self.project_slug = project_slug
        self.project_service = project_service

        self.dataset_template_slug = dataset_template_slug

        self.input_filepath = input_filepath

        self.csv_fieldnames = csv_fieldnames

        self.errors = dict()

    def get_map_documents(self, field_name):
        map_documents = self.rules.get(f'map_{field_name}')
        return map_documents

    def init_validation_metadata(self):
        self.validation_metadata = {
            self.time_dimension.lower(): {
                "min_value": float('+inf'),
                "max_value": float('-inf')
            }
        }

    def set_csv_regional_validation_rules(self):

        dataset_template_details = self.project_service.get_dataset_template_details(
            self.project_slug,
            self.dataset_template_slug
        )

        self.rules = dataset_template_details.get('rules')

        assert self.rules, \
            f"No dataset template rules found for dataset_template id: \
                {self.dataset_template_id}"

        self.time_dimension = self.rules['root_schema_declarations']['time_dimension']
        self.value_dimension = self.rules['root_schema_declarations']['value_dimension']

        self.region_dimension = self.rules['root_schema_declarations']['region_dimension']

        # specific to regional timeseries
        self.region_layer_dimension = self.rules['root_schema_declarations']['region_layer_map_key']

    def validate_row_data(self, row):
        try:
            jsonschema_validate(
                self.rules.get('root'),
                row
            )

        except SchemaError as schema_error:

            raise ValueError(
                f"Schema itself is not valid with template id. Template id: {self.dataset_template_id}. Original exception: {str(schema_error)}"
            )
        except ValidationError as validation_error:
            raise ValueError(
                f"Invalid data. Template id: {self.dataset_template_id}. Data: {str(validation_error)}. Original exception: {str(validation_error)}"
            )

        for key in self.rules['root']['properties']:

            if key == self.time_dimension.lower():
                if float(row[key]) < self.validation_metadata[
                    self.time_dimension.lower()
                ]["min_value"]:
                    self.validation_metadata[
                        self.time_dimension.lower()
                    ]["min_value"] = float(row[key])

                if float(row[key]) > self.validation_metadata[
                    self.time_dimension.lower()
                ]["max_value"]:
                    self.validation_metadata[
                        self.time_dimension.lower()
                    ]["max_value"] = float(row[key])

                continue

            if key == self.value_dimension:
                continue

            map_documents = self.get_map_documents(key)

            if map_documents:
                if row[key] not in map_documents:
                    raise ValueError(f"'{row[key]}' must be one of {map_documents.keys()}")

            if self.validation_metadata.get(key):
                if len(self.validation_metadata[key]) <= 200:  # limit harvest
                    self.validation_metadata[key].add(row[key])

            else:
                self.validation_metadata[key] = {row[key]}

        extra_template_validators = self.rules.get('template_validators')

        if extra_template_validators and extra_template_validators != 'not defined':

            for row_key in extra_template_validators.keys():
                lhs = row[row_key]

                condition_object = extra_template_validators[row_key]

                for condition in condition_object.keys():

                    rhs_value_pointer = condition_object[condition]

                    rhs = None
                    for pointer in rhs_value_pointer:
                        if pointer.startswith('&'):
                            rhs = self.rules[pointer[1:]]
                        elif pointer.startswith('{') and pointer.endswith('}'):
                            rhs = rhs[row[pointer[1:-1]]]

                        else:
                            rhs = rhs[pointer]

                    if condition == 'value_equals':
                        if lhs.lower() != rhs.lower():
                            raise ValueError(
                                f'{lhs} in {row_key} column must be equal to {rhs}.'
                            )

                    if condition == 'is_subset_of_map':
                        if not lhs.lower() in rhs.lower():
                            raise ValueError(
                                f'{lhs} in {row_key} column must be member of {rhs}.'
                            )

        return row

    def get_validated_rows(self, filepath):
        with open(filepath) as csvfile:
            # with open(self.temp_downloaded_filepath) as csvfile:
            reader = csv.DictReader(
                lower_rows(csvfile),
                fieldnames=self.csv_fieldnames,
                restkey='restkeys',
                restval='restvals'
            )

            for row in reader:
                row.pop('restkeys', None)
                row.pop('restvals', None)

                try:
                    row = self.validate_row_data(row)
                    row[self.region_layer_dimension] = \
                    self.rules[f'map_{self.region_dimension}'][row[self.region_dimension]]['region_layer']
                except Exception as err:
                    if len(self.errors) <= 50:
                        self.errors[str(err)] = str(row)

                yield row

    def __call__(self):

        self.set_csv_regional_validation_rules()
        self.init_validation_metadata()

        with Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                transient=True,
        ) as progress:
            progress.add_task(description="[bold cyan]Validating dataset. Please wait...", total=None)

            for row in self.get_validated_rows(self.input_filepath):
                pass

        if self.errors:
            for key in self.errors:
                print(f"[bold red]Invalid data:[/bold red] {self.errors[key]}")
                print(f"[bold red]Error:[/bold red] {key}")
            print("[bold red]Data is not valid against selected template. ⚠️⚠️")
            typer.Exit(1)
        else:
            print("[bold green]Data validated against selected template.[/bold green] 🎉🎉")
            typer.Exit(0)



import os
import zipfile
import hashlib
import shutil
import tempfile
from typing import Optional, Dict, List

import requests
from functools import lru_cache
from pydantic import BaseModel, model_validator
from accli.AcceleratorTerminalCliProjectService import AcceleratorTerminalCliProjectService
from accli.token import (
    get_token, get_server_url,
    get_project_slug
)

ACCLI_DEBUG = os.environ.get('ACCLI_DEBUG', False)

FOLDER_JOB_REPO_URL = 'https://github.com/IIASA-Accelerator/wkube-job.git'


def compress_folder(folder_path, output_path):
    # Fixed timestamp for normalization
    fixed_time = (1980, 1, 1, 0, 0, 0)

    with zipfile.ZipFile(output_path, 'w', zipfile.ZIP_DEFLATED) as zipf:
        for root, dirs, files in os.walk(folder_path):
            dirs.sort()
            files.sort()
            for file in files:
                file_path = os.path.join(root, file)
                arc_name = os.path.relpath(file_path, start=folder_path)
                # Add a file to zip with a fixed timestamp
                info = zipfile.ZipInfo(arc_name)
                info.date_time = fixed_time
                info.external_attr = 0o600 << 16  # Set file permissions
                with open(file_path, 'rb') as f:
                    zipf.writestr(info, f.read(), zipfile.ZIP_DEFLATED)


def get_file_sha1(file_path):
    sha1_hash = hashlib.sha1()
    with open(file_path, "rb") as f:
        chunk = f.read(8192)
        while chunk:
            sha1_hash.update(chunk)
            chunk = f.read(8192)
    return sha1_hash.hexdigest()


def copy_tree(src, dst):
    """Recursively copy from src to dst, excluding .git folders."""
    for item in os.listdir(src):
        src_path = os.path.join(src, item)
        dst_path = os.path.join(dst, item)

        if os.path.isdir(src_path):
            if item == '.git':
                continue
            os.makedirs(dst_path, exist_ok=True)
            copy_tree(src_path, dst_path)
        else:
            shutil.copy2(src_path, dst_path)


@lru_cache(maxsize=None)
def push_folder_job(directory):
    access_token = get_token()
    server_url = get_server_url()
    project_slug = get_project_slug()

    term_cli_project_service = AcceleratorTerminalCliProjectService(
        user_token=access_token,
        server_url=server_url,
        verify_cert=(not ACCLI_DEBUG)
    )

    repo_dir = tempfile.mkdtemp()
    copy_tree(directory, repo_dir)

    if os.path.isfile(f'{repo_dir}/wkube.py'):
        os.remove(f'{repo_dir}/wkube.py')

    temp_zip_path = f"{repo_dir}/temp.zip"
    compress_folder(repo_dir, temp_zip_path)

    sha256_hash = get_file_sha1(temp_zip_path)
    final_zip_path = f"{repo_dir}/{sha256_hash}.zip"
    os.rename(temp_zip_path, final_zip_path)

    presigned_push_url = term_cli_project_service.get_jobstore_push_url(
        project_slug, f"{sha256_hash}.zip"
    )

    if presigned_push_url:
        with open(final_zip_path, 'rb') as f:
            res = requests.put(
                presigned_push_url,
                data=f,
                verify=False,
            )
            res.raise_for_status()

    shutil.rmtree(repo_dir)
    return f"s3accjobstore://{sha256_hash}.zip", sha256_hash


class JobDispatchModel(BaseModel):
    is_holder_job: bool = True
    build_only_task: bool = False

    name: str

    execute_cluster: str
    job_location: str
    job_args: list
    job_kwargs: dict

    required_cores: Optional[float] = None
    required_ram: Optional[float] = None
    required_storage_local: Optional[float] = None

    # is ignored if it is a callback and child of non-free node jobs
    required_storage_workflow: Optional[float] = None

    job_secrets: Optional[dict] = {}

    timeout: Optional[int] = None
    pvc_id: Optional[str] = None
    node_id: Optional[str] = None

    ignore_duplicate_job: bool = False
    free_node: bool = False  # Only applies to immediate children jobs

    children: List['JobDispatchModel'] = []
    callback: Optional['JobDispatchModel'] = None

    def model_dump(self, *args, **kwargs):
        result = super().model_dump(*args, **kwargs)
        if result['is_holder_job'] and result['execute_cluster'] == 'WKUBE':
            if len(result['children']) > 1:
                if not result['children'][0]['job_kwargs'].get('docker_image'):
                    builder_task = result['children'][0].copy()

                    builder_task['build_only_task'] = True

                    builder_task['callback'] = result.copy()
                    builder_task['name'] = f"{builder_task['callback']['name']} -- Image Builder"
                    builder_task['callback']['name'] = f"{builder_task['callback']['name']} -- Holder"
                    return builder_task
        return result


class WKubeTaskMeta(BaseModel):
    required_cores: float
    required_ram: float
    required_storage_local: float

    # is ignored if it is a callback and child of non-free node jobs
    required_storage_workflow: float

    job_secrets: Optional[dict] = {}

    timeout: int


class WKubeTaskKwargs(BaseModel):
    docker_image: Optional[str] = None

    job_folder: str = './'

    repo_url: Optional[str] = None  # required when docker image is not present
    repo_branch: Optional[str] = None  # required when docker image is not present

    docker_filename: Optional[str] = None  # when not docker image;
    base_stack: Optional[str] = None  # when not github docker file #TODO add enum class of available base stack

    force_build: bool = False

    command: str  # may not be present with docker_image # TODO wrap a command in custom script to implement timeout or possibly log ingestion if required.

    conf: Dict[str, str] = {}

    build_timeout: Optional[int] = None

    def model_dump(self, *args, **kwargs):
        result = super().model_dump(*args, **kwargs)
        if 'job_folder' in result:
            del result['job_folder']
        return result

    @model_validator(mode="before")
    @classmethod
    def validate_root(cls, values):
        if not values.get('docker_image'):
            job_folder = values.get('job_folder', './')

            if not (values.get('repo_url') and values.get('repo_branch')):
                remote_url, branch_name = push_folder_job(
                    os.path.abspath(job_folder)
                )
                values['repo_url'] = remote_url
                values['repo_branch'] = branch_name
            else:
                if not values.get('repo_url'):
                    # TODO if has no repo url just set it
                    raise ValueError("repo_url is required")

                if not values.get('repo_branch'):
                    raise ValueError('repo_branch is required')

            if not values.get('docker_filename'):
                if not values.get("base_stack"):
                    raise ValueError("base_stack is required when dockerfile is not defined")
        return values


class WKubeTaskPydantic(WKubeTaskMeta, WKubeTaskKwargs):
    pass


class GenericTask:
    def __init__(self, *args, **kwargs):
        self.dispatch_model_task: JobDispatchModel

    def add_child(self, task):
        if self.__class__ != task.__class__:
            raise ValueError(f"task should of {self.__class__} class")
        self.dispatch_model_task.children.append(task.dispatch_model_task)

    def add_callback(self, task):
        if self.__class__ != task.__class__:
            raise ValueError(f"task should of {self.__class__} class")
        self.dispatch_model_task.callback = task.dispatch_model_task

    @property
    def description(self):
        return self.dispatch_model_task.model_dump()


class WKubeTask(GenericTask):
    def __init__(self, *t_args, **t_kwargs):
        super().__init__(*t_args, **t_kwargs)
        wkube_task_kwargs = None
        wkube_task_meta = {}

        name = t_kwargs.pop('name') if 'name' in t_kwargs else None
        if t_args or t_kwargs:
            WKubeTaskPydantic(*t_args, **t_kwargs)
            wkube_task_kwargs = WKubeTaskKwargs(*t_args, **t_kwargs)
            wkube_task_meta.update(WKubeTaskMeta(*t_args, **t_kwargs).model_dump(exclude_unset=True))

        self.dispatch_model_task = JobDispatchModel(
            name=name,
            is_holder_job=not (t_args or t_kwargs),
            execute_cluster='WKUBE',
            job_location='acc_native_jobs.dispatch_wkube_task',
            job_args=[],
            job_kwargs=wkube_task_kwargs.model_dump(exclude_unset=True) if wkube_task_kwargs else {},
            **wkube_task_meta
        )
